package rover;
import java.util.*;

public class PlanParser {

  String input;
  StringTokenizer tokens;
  String currentToken;

  PlanParser(String ip) {
    input = ip + " end";
    tokens = new StringTokenizer(input, " \t\n\r\f()+,", true);
  }
  
  Node parsePlan() throws ParseException{
  	advance();
  	Node ret = parseNode();
  	if (!currentToken.equals("end"))
  	  throw new ParseException("Extra input");
  	return ret;
  }

  Node parseNode() throws ParseException {
  	//assume we are at the ( 
    Node retNode;
    if (!currentToken.equals("("))
      throw new ParseException("Node starts with \"(\"");
    advance(); //read task or block
    if (currentToken.equals("task"))
      retNode = parseTask();
    else if (currentToken.equals("block"))
      retNode = parseBlock();
    else
      throw new ParseException("Expected keyword \"task\" or \"block\"");
    if (!currentToken.equals(")"))
      throw new ParseException("Node ends with \")\"");
    advance();
    return retNode;
  }

  Node parseTask() throws ParseException {
    ListOfConditions startConditions = null;
    ListOfConditions endConditions = null;
    ContinueType continueOnFailure = new ContinueType(false);
    Symbol taskActionName = null;
    String taskParameters = null;
    boolean succeed = true;
    long duration = 0;

    advance(); //read :id
    if (!currentToken.equals(":id"))
      throw new ParseException("Task must have :id field");
    advance(); //read task name
    Symbol taskName = new Symbol(currentToken);
    advance(); //read each node and task field

    while (currentToken.startsWith(":")) {
      if (currentToken.equals(":start-conditions")) {
        startConditions = parseConditions();
      } else if (currentToken.equals(":end-conditions")) {
        endConditions = parseConditions();
      } else if (currentToken.equals(":continue-on-failure")) {
        advance();
        continueOnFailure = new ContinueType(true);
      } else if (currentToken.equals(":action")) {
        advance();
        taskActionName = new Symbol(currentToken);
        advance();
      } else if (currentToken.equals(":fail")) {
        advance();
        succeed = false;
      } else if (currentToken.equals(":duration")) {
        advance();
        try{
        	duration = (long) Integer.parseInt(currentToken);
        }
        catch (Exception e){
        	throw new ParseException("Non numeric :duration value");
        }
        advance();
      } else
        throw new ParseException("Unrecognized node/task attribute");
    }
    Node node =
      new Task(
        taskName,
        startConditions,
        endConditions,
        continueOnFailure,
        taskActionName,
        succeed,
        duration);
    node.compileStartConditions();
    node.compileEndConditions();
    
    return node; //node
    
  }

  Node parseBlock() throws ParseException {
    ListOfConditions startConditions = null;
    ListOfConditions endConditions = null;
    ContinueType continueOnFailure = new ContinueType(false);
    ListOfNodes blockNodeList = null;
    advance(); //read :id
    if (!currentToken.equals(":id"))
      throw new ParseException("Task must have :id field");
    advance(); //read task name
    Symbol blockName = new Symbol(currentToken);
    advance(); //read each node and task field	

    while (currentToken.startsWith(":")) {
      if (currentToken.equals(":start-conditions")) {
        startConditions = parseConditions();
      } else if (currentToken.equals(":end-conditions")) {
        endConditions = parseConditions();
      } else if (currentToken.equals(":continue-on-failure")) {
        advance();
        continueOnFailure = new ContinueType(true);
      } else if (currentToken.equals(":node-list")) {
        advance();
        if (!currentToken.equals("("))
          throw new ParseException("Block node list must start with \"(\"");
        advance();
        blockNodeList = new ListOfNodes();
        while (!currentToken.equals(")")) {
          Node elem = parseNode();
          blockNodeList.pushEnd(elem);
        }
        advance(); //past )
      } else
        throw new ParseException("Unrecognized node attributes");
    }
    Node ret = new Block(
      blockName,
      blockNodeList,
      startConditions,
      endConditions,
      continueOnFailure);
    ret.compileStartConditions();
    ret.compileEndConditions();
    return ret;
  }

  void advance() throws ParseException {
    do {
      if (!tokens.hasMoreTokens())
        throw new ParseException("Incomplete input");
      currentToken = tokens.nextToken();
    } while (Character.isWhitespace(currentToken.charAt(0)));
  }

  ListOfConditions parseConditions() throws ParseException {
    advance();
    if (!currentToken.equals("("))
      throw new ParseException("Missing \"(\"");
    advance();
    if (!currentToken.equals("time"))
      throw new ParseException("Missing time condition");
    ListOfConditions loc = new ListOfConditions();
    advance();
    TimeDate td1 = parseTimeDate();
    TimeDate td2 = parseTimeDate();
    if (!currentToken.equals(")"))
      throw new ParseException("Missing \")\"");
    advance();
    loc.pushEnd(new TimeCondition(td1, td2));
    return loc;
  }

  TimeDate parseTimeDate() throws ParseException {
    //assume we are already at first token 
    boolean rel = false;
    if (currentToken.equals("+")) {
      rel = true;
      advance(); //past the +
    }
    long sec;
    try {
      sec = (long) (Integer.parseInt(currentToken));
    } catch (Exception e) {
      throw new ParseException("Bad time specifier");
    }
    advance(); //past the seconds, perhaps past the TimeDate
    int millis = 0;
    if (currentToken.equals(",")) {
      advance(); //past the .
      try {
        millis = Integer.parseInt(currentToken);
      } catch (Exception e) {
        throw new ParseException("Bad time specifier");
      }
      advance();
    }
    return new TimeDate(sec, millis, rel);
  }

}